import torch
from torch import nn
from torchvision.models.utils import load_state_dict_from_url
from nn_modules import Blind, UnBlind
from collections import OrderedDict


__all__ = ['MobileNetV2', 'mobilenet_v2']


model_urls = {
    'mobilenet_v2': 'https://download.pytorch.org/models/mobilenet_v2-b0353104.pth',
}


def _make_divisible(v, divisor, min_value=None):
    """
    This function is taken from the original tf repo.
    It ensures that all layers have a channel number that is divisible by 8
    It can be seen here:
    https://github.com/tensorflow/models/blob/master/research/slim/nets/mobilenet/mobilenet.py
    :param v:
    :param divisor:
    :param min_value:
    :return:
    """
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v


class ConvBNReLU(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size=3, stride=1, groups=1):
        super(ConvBNReLU, self).__init__();
        padding = (kernel_size - 1) // 2
        
        self.conv2d = nn.Sequential(nn.Conv2d(in_planes, out_planes, kernel_size, stride, padding, groups=groups, bias=False))
        self.norm = nn.BatchNorm2d(out_planes)
        self.relu = nn.ReLU6(inplace=False)
        self.use_noise = False
    
    def addNoise(self, bm, um, gm, igm):
        new_list = []
        for idx, layer in enumerate(self.conv2d):
            if idx == 0:
                new_list.append(Blind(bm, igm))
            new_list.append(layer)

            if idx == 0:
                new_list.append(UnBlind(um, gm, bias_size=0, use_bias=False))
        retval = nn.Sequential(*new_list)
        self.use_noise = True
        self.conv2d = retval

    def forward(self, x):
        x = self.conv2d(x)
        partition_point = x.size(0) // 3 * 2

        if self.use_noise:
            x2 = self.norm(x[0:partition_point])
            x = torch.cat((x2, x[partition_point:]), axis=0)

        else:
            x = self.norm(x)

        x = self.relu(x)
        return x

class ConvBN(nn.Module):
    def __init__(self, in_planes, out_planes, kernel_size=3, stride=1, padding=1, groups=1):
        super(ConvBN, self).__init__();
        
        self.conv2d = nn.Sequential(nn.Conv2d(in_planes, out_planes, kernel_size, stride, padding, groups=groups, bias=False))
        self.norm = nn.BatchNorm2d(out_planes)
        self.use_noise = False
    
    def addNoise(self, bm, um, gm, igm):
        new_list = []
        for idx, layer in enumerate(self.conv2d):
            if idx == 0:
                new_list.append(Blind(bm, igm))
            new_list.append(layer)

            if idx == 0:
                new_list.append(UnBlind(um, gm, bias_size=0, use_bias=False))
        retval = nn.Sequential(*new_list)
        self.use_noise = True
        self.conv2d = retval

    def forward(self, x):
        x = self.conv2d(x)
        partition_point = x.size(0) // 3 * 2

        if self.use_noise:
            x2 = self.norm(x[0:partition_point])
            x = torch.cat((x2, x[partition_point:]), axis=0)

        else:
            x = self.norm(x)

        return x


class InvertedResidual(nn.Module):
    def __init__(self, inp, oup, stride, expand_ratio):
        super(InvertedResidual, self).__init__()
        self.stride = stride
        assert stride in [1, 2]

        hidden_dim = int(round(inp * expand_ratio))
        self.use_res_connect = self.stride == 1 and inp == oup

        layers = []
        if expand_ratio != 1:
            # pw
            layers.append(ConvBNReLU(inp, hidden_dim, kernel_size=1))
        layers.extend([
            # dw
            ConvBNReLU(hidden_dim, hidden_dim, stride=stride, groups=hidden_dim),
            # pw-linear
            ConvBN(hidden_dim, oup, kernel_size=1, stride=1, padding=0)
        ])
        self.conv = nn.Sequential(*layers)

    def forward(self, x):
        if self.use_res_connect:
            return x + self.conv(x)
        else:
             return self.conv(x)


    def addNoise(self, bm, um, gm, igm):
        for layer in self.conv:
            layer.addNoise(bm, um, gm, igm)

class MobileNetV2(nn.Module):
    def __init__(self,
                 num_classes=1000,
                 width_mult=1.0,
                 inverted_residual_setting=None,
                 round_nearest=8,
                 block=None,
                 device=None,
                 dtype=None):
        """
        MobileNet V2 main class
        Args:
            num_classes (int): Number of classes
            width_mult (float): Width multiplier - adjusts number of channels in each layer by this amount
            inverted_residual_setting: Network structure
            round_nearest (int): Round the number of channels in each layer to be a multiple of this number
            Set to 1 to turn off rounding
            block: Module specifying inverted residual building block for mobilenet
        """
        super(MobileNetV2, self).__init__()

        if block is None:
            block = InvertedResidual
        input_channel = 32
        last_channel = 1280

        self.flattened = []
        if inverted_residual_setting is None:
            inverted_residual_setting = [
                # t, c, n, s
                [1, 16, 1, 1],
                [6, 24, 2, 2],
                [6, 32, 3, 2],
                [6, 64, 4, 2],
                [6, 96, 3, 1],
                [6, 160, 3, 2],
                [6, 320, 1, 1],
            ]

        self.device = device
        self.dtype  = dtype
        # only check the first element, assuming user knows t,c,n,s are required
        if len(inverted_residual_setting) == 0 or len(inverted_residual_setting[0]) != 4:
            raise ValueError("inverted_residual_setting should be non-empty "
                             "or a 4-element list, got {}".format(inverted_residual_setting))

        # building first layer
        input_channel = _make_divisible(input_channel * width_mult, round_nearest)
        self.last_channel = _make_divisible(last_channel * max(1.0, width_mult), round_nearest)
        features = [ConvBNReLU(3, input_channel, stride=2)]
        # building inverted residual blocks
        for t, c, n, s in inverted_residual_setting:
            output_channel = _make_divisible(c * width_mult, round_nearest)
            for i in range(n):
                stride = s if i == 0 else 1
                features.append(block(input_channel, output_channel, stride, expand_ratio=t))
                input_channel = output_channel
        # building last several layers
        features.append(ConvBNReLU(input_channel, self.last_channel, kernel_size=1))
        # make it nn.Sequential
        self.features = nn.Sequential(*features)

        # building classifier
        self.classifier = nn.Sequential(
            nn.Dropout(0.2),
            nn.Linear(self.last_channel, num_classes),
        )

        # weight initialization
        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out')
                if m.bias is not None:
                    nn.init.zeros_(m.bias)
            elif isinstance(m, nn.BatchNorm2d):
                nn.init.ones_(m.weight)
                nn.init.zeros_(m.bias)
            elif isinstance(m, nn.Linear):
                nn.init.normal_(m.weight, 0, 0.01)
                nn.init.zeros_(m.bias)

    def _forward_impl(self, x):
        # This exists since TorchScript doesn't support inheritance, so the superclass method
        # (this one) needs to have a name other than `forward` that can be accessed in a subclass
        x = self.features(x)
        # Cannot use "squeeze" as batch-size can be 1 => must use reshape with x.shape[0]
        x = nn.functional.adaptive_avg_pool2d(x, 1).reshape(x.shape[0], -1)
        x = self.classifier(x)
        return x

    def forward(self, x):
        return self._forward_impl(x)


    def addNoise(self, bm, um, gm, igm):
        for l  in self.features:
            l.addNoise(bm, um, gm, igm)

        print("------------------------")
        new_classifier_list = []
        for c in self.classifier:
            if isinstance(c, nn.Linear):
                new_classifier_list.append(Blind(bm, igm))
                new_layer = nn.Linear(c.in_features, c.out_features, bias=False)
                new_layer.weight = c.weight
                new_classifier_list.append(new_layer)
                ub = UnBlind(um, gm, c.out_features, use_bias=True)
                ub.bias = c.bias
                new_classifier_list.append(ub)
            else:
                new_classifier_list.append(c)
        self.classifier = nn.Sequential(*new_classifier_list)

    def copy_state_dict(self, state_dict):
        self_state_dict = self.state_dict()
        new_state_dict = OrderedDict()
        iter_old = iter(state_dict)
        iter_self= iter(self_state_dict)
        print(len(state_dict), len(self_state_dict))
        if (len(state_dict) != len(self_state_dict)):
            print("Miss match state dict")
            assert(False)
        #print(len(state_dict), len(self_state_dict))
        for _ in range(len(state_dict)):
            key = next(iter_self)
            old_key = next(iter_old)
            #print(key, old_key)
            val = state_dict[old_key]
            new_state_dict[key] = val

        self.load_state_dict(new_state_dict)

def mobilenet_v2(pretrained=False, progress=True, **kwargs):
    """
    Constructs a MobileNetV2 architecture from
    `"MobileNetV2: Inverted Residuals and Linear Bottlenecks" <https://arxiv.org/abs/1801.04381>`_.
    Args:
        pretrained (bool): If True, returns a model pre-trained on ImageNet
        progress (bool): If True, displays a progress bar of the download to stderr
    """
    model = MobileNetV2(**kwargs)
    if pretrained:
        state_dict = load_state_dict_from_url(model_urls['mobilenet_v2'],
                                              progress=progress)
        model.copy_state_dict(state_dict)
    return model